from __future__ import annotations

from typing import List, Generic, TypeVar, Callable, Union, Any, Type, Optional

from .monad import Functor, Either


TokT = TypeVar('TokT')
E = TypeVar('E')
class Reader(Generic[TokT, E]):

    def __init__(self, content: List[TokT]):
        self.content = content
        self.counter = 0

    def expect(
            self,
            predicate: Callable[[TokT], bool],
            error: E) -> Either[E, TokT]:

        token = self.content[self.counter]
        self.counter += 1

        if predicate(token):
            return Either.pure(token)

        return Either.fail(error)

    def test(self, predicate: Callable[[TokT], bool]) -> bool:
        return predicate(self.content[self.counter])

    def load_counter(self, update: int) -> None:
        self.counter = update

    def get_counter(self) -> int:
        return self.counter


S = TypeVar('S')
T = TypeVar('T')
class Parser(Generic[TokT, S, E], Functor[S]):
    """
    An interface for all parser combinators.
    """

    def __init__(self, f: Callable[[Reader[TokT, E]], Either[E, List[S]]]):
        self.func = f

    def parse(self, reader: Reader[TokT, E]) -> Either[E, List[S]]:
        return self.func(reader)

    def __add__(self, p: Parser[TokT, S, E]) -> Parser[TokT, S, E]:
        return Join(self, p)

    def __or__(self, p: Parser[TokT, S, E]) -> Parser[TokT, S, E]:
        return Or(self, p)

    def fmap(self, f: Callable[[S], T]) -> Parser[TokT, T, E]:
        return FunctorialParser(f, self)


U = TypeVar('U')
class FunctorialParser(Parser[TokT, S, E]):

    def __init__(self, f: Callable[[U], S], a: Parser[TokT, U, E]):
        self.f = f
        self.inner = a

    def parse(self, reader: Reader[TokT, E]) -> Either[E, List[S]]:
        return self.inner.parse(reader).fmap( # type: ignore
            lambda a: [self.f(i) for i in a]
        )


class Join(Parser[TokT, S, E]):

    def __init__(self, a: Parser[TokT, S, E], b: Parser[TokT, S, E]):
        self.a = a
        self.b = b

    def parse(self, reader: Reader[TokT, E]) -> Either[E, List[S]]:
        # mypy doesn't support lambdas...

        def outer(a: List[S]):

            def inner(b: List[S]):
                return a + b

            return self.b.parse(reader).fmap(inner) # type: ignore

        return self.a.parse(reader).fmap(outer) # type: ignore


class Or(Parser[TokT, S, E]):

    def __init__(self, a: Parser[TokT, S, E], b: Parser[TokT, S, E]):
        self.a = a
        self.b = b

    def parse(self, reader: Reader[TokT, E]) -> Either[E, List[S]]:
        counter = reader.get_counter()

        result = self.a.parse(reader)
        if result.success():
            return result

        reader.load_counter(counter)
        return self.b.parse(reader)


from typing import List, Generic, TypeVar, Callable, Union, Any, Type, Optional

from .combinator import Reader, Parser
from .monad import Either


S = TypeVar('S')
TokT = TypeVar('TokT')
def satisfy(predicate: Callable[[TokT], bool]) -> Parser[TokT, S, str]:

    def inner(reader: Reader[TokT, str]) -> Either[str, List[S]]:
        return reader.expect( # type: ignore
            predicate, f'Parsing failed at token {reader.get_counter() - 1}'
        ).fmap(lambda a: [a])

    return Parser(inner)


E = TypeVar('E')
def optional(parser: Parser[TokT, S, E]) -> Parser[TokT, S, E]:

    def inner(reader: Reader[TokT, E]) -> Either[E, List[S]]:
        counter = reader.get_counter()

        result = parser.parse(reader)
        if result.success():
            return result

        reader.load_counter(counter)
        return Either.pure([])

    return Parser(inner)


def zeroOrMore(parser: Parser[TokT, S, E]) -> Parser[TokT, S, E]:

    def outer(reader: Reader[TokT, E]) -> Either[E, List[S]]:
        out: Either[E, List[S]] = Either.pure([])
        counter = reader.get_counter()
        parse = parser.parse(reader)

        def middle(a: List[S]) -> List[S]:

            def inner(b: List[S]) -> List[S]:
                return a + b

            return parse.fmap(inner) # type: ignore

        while parse.success():
            out = out.fmap(middle) # type: ignore
            counter = reader.get_counter()

        reader.load_counter(counter)

        return out

    return Parser(outer)


def oneOrMore(parser: Parser[TokT, S, E]) -> Parser[TokT, S, E]:
    return parser + zeroOrMore(parser)


def runParser(parser: Parser[TokT, S, E], tokens: List[TokT]) -> List[S]:

    def raise_error(error: E) -> None:
        raise ValueError(error)

    reader = Reader[TokT, E](tokens)
    result = parser.parse(reader)

    result.on_fail(raise_error)

    return result.item()


from typing import List, Generic, TypeVar, Callable, Optional, Union, Type, Any


F = TypeVar('F')
T = TypeVar('T')
class Functor(Generic[F]):

    def fmap(self, f: Callable[[F], T]) -> Functor[T]:
        raise NotImplementedError


class Monad(Functor[F]):

    @staticmethod
    def join(m: Monad[Monad[F]]) -> Monad[F]:
        return m.bind(lambda x: x)

    @classmethod
    def pure(cls: Type[Monad[F]], a: F) -> Monad[F]:
        if cls == Monad:
            raise NotImplementedError

        return cls.pure(a)

    def bind(self, f: Callable[[F], Monad[T]]) -> Monad[T]:
        raise NotImplementedError


ErrorT = TypeVar('ErrorT')
class Either(Generic[ErrorT, F], Monad[F]):

    @staticmethod
    def fail(error: ErrorT) -> Either[ErrorT, F]:
        return Left(error)

    @classmethod
    def pure(cls: Type[Either[ErrorT, F]], a: F) -> Either[ErrorT, F]:
        return Right(a)

    def success(self) -> bool:
        raise NotImplementedError

    def on_fail(self, f: Callable[[ErrorT], None]) -> None:
        raise NotImplementedError

    def item(self) -> F:
        raise NotImplementedError


class Left(Either[ErrorT, F]):

    def __init__(self, error: ErrorT):
        self.left = error

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Left):
            return False

        return self.left == other.left

    def fmap(self, f: Callable[[F], T]) -> Either[ErrorT, T]:
        return Either.fail(self.left)

    def bind(self, f: Callable[[F], Any]) -> Either[ErrorT, T]:
        return Either.fail(self.left)

    def success(self) -> bool:
        return False

    def on_fail(self, f: Callable[[ErrorT], None]) -> None:
        f(self.left)

    def item(self) -> F:
        raise ValueError('Left does not yields an item')


class Right(Either[ErrorT, F]):

    def __init__(self, value: F):
        self.right = value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Right):
            return False

        return self.right == other.right

    def fmap(self, f: Callable[[F], T]) -> Either[ErrorT, T]:
        return Either.pure(f(self.right))

    def bind(self, f: Callable[[F], Any]) -> Either[ErrorT, T]:
        return f(self.right)

    def success(self) -> bool:
        return True

    def on_fail(self, f: Callable[[ErrorT], None]) -> None:
        pass


from .combinator import Parser

from .util import (
    Parser,
    runParser,
    satisfy,
    optional,
    zeroOrMore,
    oneOrMore
)


__all__ = [
    'Parser',
    'runParser',
    'satisfy',
    'optional',
    'zeroOrMore',
    'oneOrMore'
]


import pathlib
from setuptools import setup, find_packages


PATH = pathlib.Path(__file__).parent
README = (PATH / 'README.md').read_text()

setup(
    name='x-nlp',
    version='0.1.0',
    description='Library of helper functions for Natural Language Processing '
                'related algorithms (designed for the course).',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/techx-nlp/x-nlp',
    author='David Ma',
    author_email='davidma@davidma.cn',
    license='MIT',
    packages=find_packages()
)